/*
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.jambalaya.ProtegeDataBean;

import java.io.Serializable;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Vector;

import javax.swing.Icon;

import ca.uvic.csr.shrimp.AttributeConstants;
import ca.uvic.csr.shrimp.DataBean.AbstractDataBean;
import ca.uvic.csr.shrimp.DataBean.Artifact;
import ca.uvic.csr.shrimp.DataBean.Relationship;
import ca.uvic.csr.shrimp.ShrimpApplication.AbstractShrimpApplication;
import ca.uvic.csr.shrimp.ShrimpApplication.ApplicationAccessor;
import ca.uvic.csr.shrimp.gui.ProgressDialog;
import ca.uvic.csr.shrimp.util.CollectionUtils;
import edu.stanford.smi.protege.model.Cls;
import edu.stanford.smi.protege.model.Facet;
import edu.stanford.smi.protege.model.Frame;
import edu.stanford.smi.protege.model.Instance;
import edu.stanford.smi.protege.model.KnowledgeBase;
import edu.stanford.smi.protege.model.Model;
import edu.stanford.smi.protege.model.Project;
import edu.stanford.smi.protege.model.Reference;
import edu.stanford.smi.protege.model.SimpleInstance;
import edu.stanford.smi.protege.model.Slot;
import edu.stanford.smi.protege.model.ValueType;
import edu.stanford.smi.protege.widget.ClsWidget;

/**
 * Protege Data Bean: This is a storage unit for the Protege-Shrimp system. It
 * is designed to hold and maintain the data structure being used by the system.
 * 
 * @author Casey Best, Rob Lintern, Nasir Rather
 */

public class ProtegeDataBean extends AbstractDataBean implements Serializable {

    protected static final String UNKNOWN_ART_TYPE = "unknown";
    public static final String INSTANCE_ART_TYPE = "Instance";
    public static final String CLASS_ART_TYPE = "Class";
    public static final String META_CLASS_ART_TYPE = "Meta Class";
    public static final String SYSTEM_CLASS_ART_TYPE = "System Class";
    public static final String SYSTEM_META_CLASS_ART_TYPE = "System Meta Class";
    public static final String DIRECT_SUBCLASS_SLOT_TYPE = "has subclass";
    public static final String DIRECT_INSTANCE_SLOT_TYPE = "has instance";
    
    public static final String GROUP_NAME_CLASS_INSTANCE_HIERARCHY = "Class-Instance Hierarchy";
    public static final String GROUP_NAME_SLOT_INSTANCE = "Slots Connecting Instances";
    //public static final String GROUP_NAME_SLOT_TEMPLATE_VALUE = "(Default Value) Group";
    public static final String GROUP_NAME_SLOT_TEMPLATE_ALLOWED_CLASS = "Slots Connecting Classes";
    public static final String GROUP_NAME_SYSTEM = "System";
    public static final String GROUP_NAME_REIFIED_RELS = "Direct Binary Relationships";
    
    public static final String SUFFIX_SLOT_INSTANCE = "";
    public static final String SUFFIX_SLOT_REIFIED_REL = "";
    public static final String SUFFIX_SLOT_TEMPLATE_ALLOWED_CLASS = " (Connects Classes)";
    public static final String SUFFIX_SLOT_TEMPLATE_VALUE = " (Default Value)";
    protected static final String[] SUBCLASS_OF_HIERARCHY = new String[] { DIRECT_SUBCLASS_SLOT_TYPE };
    protected static final String[] INSTANCE_OF_HIERARCHY = new String[] { DIRECT_INSTANCE_SLOT_TYPE };
    protected static final String[] SUBCLASS_OF_AND_INSTANCE_OF_HIERARCHY = new String[] { DIRECT_SUBCLASS_SLOT_TYPE, DIRECT_INSTANCE_SLOT_TYPE };
    private static final boolean DEFAULT_SHOW_INCLUDED_FRAMES = true;
    // whether or not to consider the instances (potentially thousands) of metaclasses
    //protected static boolean CONSIDER_METACLASS_INSTANCES = true;
    protected static boolean UNREIFY_RELATION_INSTANCES = true;
    //protected static String ATTR_FRAME_CREATION_TIMESTAMP = "frame creation/ timestamp";
    //protected static String ATTR_FRAME_LAST_MODIFICATION_TIMESTAMP = "frame last modification timestamp";
    //protected static String ATTR_FRAME_LAST_MODIFIER = "frame last modifier";
    //protected static String ATTR_FRAME_CREATOR = "frame creator";
    protected static String NOM_ATTR_IS_INCLUDED = "is included";
    //protected static String ATTR_VALUE_IS_INCLUDED = "included";
    //protected static String ATTR_VALUE_NOT_INCLUDED = "not included";

    /** The Protege Knowledgebase this bean wraps around */
    protected KnowledgeBase knowledgeBase;

    /** The Protege project that this bean refers to */
    protected Project project;

    /*
     * |||||||||||||||||||||| Start of buffered Data from the back-end
     * ||||||||||||||||||||||
     */

    /** Holds the widget associated with each frame, key is the Protege Frame. */
    private Map clsWidgets;

    /** Maps a protege frame to its text. */
    private static Map frameToBrowserText = new HashMap();

    /** Maps a frame to its own slots * */
    private Map frameToOwnSlots;

    /**
     * Maps a frame its own slot values Key is the frame, value is another map;
     * This second map's key is the slot and its value is a vector of own slot
     * values
     */
    private Map frameToOwnSlotValues;

    /** Maps a slot to its value type * */
    private Map slotToValueType;

    private Map artifactToOutgoingUnreifiedRels;

    private Map artifactToIncomingUnreifiedRels;

    public boolean shouldImportSlotTemplateValues = false;

    public boolean shouldImportSlotAllowedValues = true;

    protected Map relTypeToGroupMap;

    private boolean relsUnreified = false;

    protected Map frameToArtifactTypeMap;

    /*
     * |||||||||||||||||||||| End of buffered Data from the back-end
     * ||||||||||||||||||||||
     */

    /**
     * A set of root clses. These root classes, and all subclasses and instances
     * of these root classes, are considered the "working" set of this databean.
     */
    private Set rootClses;

    /**
     * A set of super classes and super types of the chosen root classes. These
     * are used for determining whether or not a slot should be imported into
     * shrimp
     */
    protected Set validSlotSuperClsesAndTypes;

    private ProtegeMetrics metrics;

    private Collection relationSubClassesAndInstances;

    private boolean showIncludedFrames = DEFAULT_SHOW_INCLUDED_FRAMES;

    /**
     * Creates a new Data Bean, initializing all of the attributes.
     */
    public ProtegeDataBean() {
        super();

        clsWidgets = new HashMap();
        frameToBrowserText = new HashMap();
        frameToOwnSlots = new HashMap();
        frameToOwnSlotValues = new HashMap();
        slotToValueType = new HashMap();
        artifactToOutgoingUnreifiedRels = new HashMap();
        artifactToIncomingUnreifiedRels = new HashMap();
        relsUnreified = false;
        relTypeToGroupMap = new HashMap();
        relationSubClassesAndInstances = null;
        rootClses = new HashSet();
        metrics = new ProtegeMetrics();
        frameToArtifactTypeMap = new HashMap();
        //clsToInstancesCount = new HashMap();
        //clsToSubClassesCount = new HashMap();

        //listen for class movements (shh, they're very quiet!) and update metrics
        knowledgeBase.addClsListener(new MetricsListener());
    }

    public ProtegeDataBean(Project project) {
    	this();
        this.project = project;
        this.knowledgeBase = project.getKnowledgeBase();
    }
    
    public ProtegeDataBean(URI projectURI) {
    	this(loadProject(projectURI));
    }
    
    private static Project loadProject(URI projectURI) {
    	Project project = null;
		try {
	    	Collection errors = new ArrayList();
			project = new Project(projectURI.toString(), errors);
			if (errors.size() != 0) {
				System.out.println("Error loading protege project: " + errors);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return project;
    }
    
    /**
     * @see ca.uvic.csr.shrimp.DataBean.DataBean#clearBufferedData()
     */
    public void clearBufferedData() {
        if (dataIsDirty()) {
            super.clearBufferedData();
            clsWidgets.clear();
            //alreadyCreatedParents.clear();
            frameToBrowserText.clear();
            frameToOwnSlots.clear();
            frameToOwnSlotValues.clear();
            slotToValueType.clear();
            artifactToOutgoingUnreifiedRels.clear();
            artifactToIncomingUnreifiedRels.clear();
            relsUnreified = false;
            metrics.clear();
            relTypeToGroupMap.clear();
            relationSubClassesAndInstances = null;
            frameToArtifactTypeMap.clear();
            //clsToInstancesCount.clear();
            //clsToSubClassesCount.clear();

        }
    }

    public ProtegeMetrics getMetrics() {
        return metrics;
    }

    /**
     * Returns the protege project this databean is accessing.
     */
    public Project getProject() {
        return project;
    }

    /**
     */
    public void setDefaultRootClses() {
        setRootClses(getDefaultRootClasses());
    }

    /**
     * @param rootFrames the root classes or instances
     */
    public void setRootClses(Set rootFrames) {
        if (!this.rootClses.equals(rootFrames)) {
            setDataIsDirty(); // TODO is the data really dirty?
            clearBufferedData();
            this.rootClses = new HashSet(rootFrames);

            // create a list of all valid slot source and destination clses
            validSlotSuperClsesAndTypes = new HashSet();
            for (Iterator iter = rootClses.iterator(); iter.hasNext();) {
            	Frame frame = (Frame) iter.next();
                validSlotSuperClsesAndTypes.addAll(getDirectSuperClsesAndDirectTypesRecursive(frame, new HashSet()));					
            }
        }
    }

    /**
     * default root classes are any subclasses of :THING that are not system
     * classes
     * 
     * @return A set of clses
     */
    public Set getDefaultRootClasses() {
        /*
         * Set defaultRootClses = new HashSet(); boolean atLeastOneValidRoot =
         * false; for (Iterator iter =
         * knowledgeBase.getRootCls().getDirectSubclasses().iterator();
         * iter.hasNext();) { Cls cls = (Cls) iter.next(); if (!cls.isSystem()) {
         * defaultRootClses.add(cls); atLeastOneValidRoot = true; } } // make
         * :THING the root, if no other roots found if (!atLeastOneValidRoot) {
         * defaultRootClses.add(knowledgeBase.getRootCls()); } return
         * defaultRootClses;
         */

        Set defaultRootClses = new HashSet();
        defaultRootClses.add(knowledgeBase.getRootCls());
        return defaultRootClses;
    }

    public Set getRootClses() {
        return rootClses;
    }

    /**
     * Turns reified relations (instances of :DIRECTED-BINARY_RELATION) into
     * relationships
     *  
     */
    private void unreifyRelationInstances() {
        if (!UNREIFY_RELATION_INSTANCES || relsUnreified) {
			return;
		}

        Cls clsBinaryRelation = knowledgeBase.getCls(":DIRECTED-BINARY-RELATION");
        if (clsBinaryRelation == null) {
			return;
		}

        Slot toSlot = knowledgeBase.getSlot(":TO");
        Slot fromSlot = knowledgeBase.getSlot(":FROM");
        for (Iterator iter = getRelationSubClassesAndInstances().iterator(); iter.hasNext();) {
            Instance relationInstance = (Instance) iter.next();
            Instance toInstance = (Instance) relationInstance.getOwnSlotValue(toSlot);
            Instance fromInstance = (Instance) relationInstance.getOwnSlotValue(fromSlot);
            if (toInstance != null && fromInstance != null) {
                Artifact toArtifact = findArtifact(toInstance);
                Artifact fromArtifact = findArtifact(fromInstance);
                if (toArtifact != null && fromArtifact != null) {
                    String relType = reifiedRelToRelType(relationInstance);
                    Relationship rel = createRel(null, fromArtifact, toArtifact, relType);
                    Collection outgoingUnreifiedRels = (Collection) artifactToOutgoingUnreifiedRels.get(fromArtifact);
                    if (outgoingUnreifiedRels == null) {
                        outgoingUnreifiedRels = new ArrayList();
                        artifactToOutgoingUnreifiedRels.put(fromArtifact, outgoingUnreifiedRels);
                    }
                    if (!outgoingUnreifiedRels.contains(rel)) {
                        outgoingUnreifiedRels.add(rel);
                    }
                    Collection incomingUnreifiedRels = (Collection) artifactToIncomingUnreifiedRels.get(toArtifact);
                    if (incomingUnreifiedRels == null) {
                        incomingUnreifiedRels = new ArrayList();
                        artifactToIncomingUnreifiedRels.put(toArtifact, incomingUnreifiedRels);
                    }
                    if (!incomingUnreifiedRels.contains(rel)) {
                        incomingUnreifiedRels.add(rel);
                    }
                }
            }
        }
        relsUnreified = true;
    }

    private String reifiedRelToRelType(Instance relationInstance) {
        return frameToArtifactName(relationInstance) + " (" + frameToArtifactName(relationInstance.getDirectType()) + ")";
    }

    private Set getDirectSuperClsesAndDirectTypesRecursive(Frame frame, Set seenSoFar) {
        Set directSuperClsesAndDirectTypes = new HashSet();

        if (seenSoFar.contains(frame)) {
            return directSuperClsesAndDirectTypes;
        }
        seenSoFar.add(frame);

        // add direct super classes of cls
        if (frame instanceof Cls) {
			Cls cls = (Cls) frame;
	        directSuperClsesAndDirectTypes.addAll(cls.getDirectSuperclasses());
	        // add direct types of cls
	        //directSuperClsesAndDirectTypes.addAll(cls.getDirectTypes()); //
	        // Protege 2.0
	        directSuperClsesAndDirectTypes.add(cls.getDirectType()); // Protege 1.x
		} else if (frame instanceof Instance) {
			Instance instance = (Instance) frame;
	        directSuperClsesAndDirectTypes.add(instance.getDirectType());
		}

        // recursively add direct super classes and direct types of these newly
        // found classes
        Set allSuperClsesAndTypes = new HashSet(directSuperClsesAndDirectTypes);
        for (Iterator iter = directSuperClsesAndDirectTypes.iterator(); iter.hasNext();) {
            Cls directSuperClassOrDirectType = (Cls) iter.next();
            allSuperClsesAndTypes.addAll(getDirectSuperClsesAndDirectTypesRecursive(directSuperClassOrDirectType, seenSoFar));
        }

        return allSuperClsesAndTypes;
    }

    /**
     * Returns the Protege Class Widget for the given frame. It is important to
     * keep all of the frames looking the same so the same widget is used in
     * each artifact representing the same frame.
     * 
     * @param frame
     *            The frame to get the widget for
     */
    ClsWidget getClsWidget(Frame frame) {
        ClsWidget widget = (ClsWidget) clsWidgets.get(frame);

        if (widget == null) {
            Instance instance = (Instance) frame;
            Project project = instance.getProject();
            widget = project.createRuntimeClsWidget(instance);
            clsWidgets.put(frame, widget);
        }

        return widget;
    }

    /**
     * Saves the project back to file
     */
    public void save() {
        Collection errors = new ArrayList();
        project.save(errors);
        if (errors.size() != 0) {
			System.out.println("Problems saving: " + errors);
		}
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.DataBean#getArtifactExternalIDFromString(java.lang.String)
     */
    public Object getArtifactExternalIDFromString(String IDInStringForm) {
        return IDInStringForm;
    }

    /*
     * (non-Javadoc)
     * 
     * @see ca.uvic.csr.shrimp.DataBean.DataBean#getStringFromExternalArtifactID(java.lang.Object)
     */
    public String getStringFromExternalArtifactID(Object externalID) {
        return externalID.toString();
    }

    /*
     * (non-Javadoc)
     * 
     * @see ca.uvic.csr.shrimp.DataBean.DataBean#getStringFromExternalRelationshipID(java.lang.Object)
     */
    public String getStringFromExternalRelationshipID(Object externalID) {
        return ""; // protege relationships dont have and external id
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.DataBean#getDefaultCprels()
     */
    public String[] getDefaultCprels() {
        String[] cprels = { DIRECT_SUBCLASS_SLOT_TYPE, DIRECT_INSTANCE_SLOT_TYPE };
        return cprels;
    }
    
    public boolean getDefaultCprelsInverted() {
        return false;
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#getArtifactsFromBackEnd()
     */
    protected Vector findAllArtifactsInBackEnd() {
        Set backEndArts = new HashSet();
        // for each root cls, find all its subclasses and all its instances
        for (Iterator iter = rootClses.iterator(); iter.hasNext() && !ProgressDialog.isCancelled();) {
            Frame rootFrame = (Frame) iter.next();
            Artifact rootArt = findArtifact(rootFrame, true, false, true);
            if (rootArt != null) {
                backEndArts.add(rootArt);
                getDescendentsRecursive(rootFrame, true, true, new HashSet(), new HashSet(), false, backEndArts);
            }
        }
        return new Vector(backEndArts);
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#findArtifactTypesInBackEnd()
     */
    protected Vector findArtifactTypesInBackEnd() {
        Vector types = new Vector();
        types.add(CLASS_ART_TYPE);
        types.add(INSTANCE_ART_TYPE);
        types.add(META_CLASS_ART_TYPE);
        if (project.getDisplayHiddenClasses()) {
            types.add(SYSTEM_META_CLASS_ART_TYPE);
            types.add(SYSTEM_CLASS_ART_TYPE);
        }
        return types;
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#findRelationshipTypesInBackEnd()
     */
    protected Vector findRelationshipTypesInBackEnd() {
        createGroupToRelTypeMappings();
        Vector types = new Vector(relTypeToGroupMap.keySet());
        return types;
    }

    /** Returns the browser text for a frame. */
    protected static String getBrowserText(Frame frame) {
        if (frame == null) {
            return "null";
        }
        String browserText = (String) frameToBrowserText.get(frame);
        if (browserText == null) {
            browserText = frame.getBrowserText();
            frameToBrowserText.put(frame, browserText);
        }
        return browserText;
    }

    /** Returns the value type of a slot. */
    protected ValueType getSlotValueType(Slot slot) {
        ValueType valueType = (ValueType) slotToValueType.get(slot);
        if (valueType == null) {
            valueType = slot.getValueType();
            slotToValueType.put(slot, valueType);
        }
        return valueType;
    }

    protected Artifact createArtifact(Frame frame) {
        ProtegeArtifact art = new ProtegeArtifact(this, frame);
        return art;
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#createArtifact(Object)
     */
    protected Artifact createArtifact(Object externalId) {
        Frame frame = knowledgeBase.getFrame((String) externalId);
        if (frame == null) {
            return null;
        }
        return createArtifact(frame);
    }

    /*
     * for Ebi private Vector ttls = new Vector (); private void
     * createTerminalToTerminalLink (String label, String type, String
     * srcNodeID, String srcTerminalID, String destNodeID, String
     * destTerminalID) { Artifact parent = findArtifactFromName(srcNodeID);
     * Artifact child = findArtifactFromName(destNodeID); String relID =
     * correctID (null); Relationship ttl = createRel (relID, parent, child,
     * type, label); ttl.setName(label); ttl.setCustomizedData
     * (Relationship.NOM_ATTR_SOURCE_TERMINAL_ID, srcTerminalID);
     * ttl.setCustomizedData (Relationship.NOM_ATTR_TARGET_TERMINAL_ID,
     * destTerminalID); ttls.add(ttl); }
     * 
     * private Artifact findArtifactFromName (String name) { Collection frames =
     * knowledgeBase.getFrames(); for (Iterator iter = frames.iterator();
     * iter.hasNext();) { Frame frame = (Frame) iter.next(); if
     * (getBrowserText(frame).equals(name)) { return
     * findArtifact(frame.getName()); } } return null; }
     * 
     * private void createTerminal(Artifact artifact, String terminalId, String
     * type) { String terminalName; String terminalType;
     * 
     * if (type.equals("in")) { terminalName = "in"; terminalType =
     * "InTerminal"; } else if (type.equals("out")) { terminalName = "out";
     * terminalType = "OutTerminal"; } else if (type.equals("fault")) {
     * terminalName = "fault"; terminalType = "FaultTerminal"; } else {
     * terminalName = "unknown"; terminalType = "UnknownTerminal"; }
     * 
     * Map terminalProperties = new HashMap ();
     * terminalProperties.put(terminalId + "_Name", terminalName);
     * terminalProperties.put(terminalId + "_Type", terminalType);
     * artifact.setCustomizedData (terminalId, terminalProperties); }
     */

    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#createEmptyRelationship(Object,
     *      java.lang.String, java.lang.String, java.util.Vector)
     */
    protected Relationship createEmptyRelationship(Object externalId, String name, String type, Vector artifacts) {
        ProtegeRelationship rel = new ProtegeRelationship(this, name, type, artifacts, externalId);
        return rel;
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#findRootArtifactsInBackEnd(java.lang.String)
     */
    protected Vector findRootArtifactsInBackEnd(String[] cprels) {
        Vector rootArts = new Vector();

        // if is the "is-a" or just the "subclass of" hierarhcy then use the
        // root clses as the roots
        if (CollectionUtils.haveSameElements(cprels, SUBCLASS_OF_AND_INSTANCE_OF_HIERARCHY) || CollectionUtils.haveSameElements(cprels, SUBCLASS_OF_HIERARCHY)) {
            for (Iterator iter = rootClses.iterator(); iter.hasNext();) {
                Frame possibleRoot = (Frame) iter.next();
                Artifact art = findArtifactByExternalId(frameToExternalArtifactID(possibleRoot));
                if (art != null) {
                    rootArts.addElement(art);
                }
            }
        } else if (cprels.length == 0) {
            rootArts.addAll(getArtifacts(true));
        } else {
            rootArts = defaultFindRootArtifactsInBackEnd(cprels);
//            // if not the "is-a" or "none" hierarchy then we get roots by
//            // looking at the values of the slots/cprels of interest
//            Collection validSlots = new ArrayList();
//            for (int i = 0; i < cprels.length; i++) {
//                String cprel = cprels[i];
//                Slot tmpSlot = relationshipTypeToSlot(cprel);
//                if (tmpSlot != null && !isRelTypeFiltered(cprel)) {
//                    validSlots.add(tmpSlot);
//                }
//            }
//            for (Iterator slotIterator = validSlots.iterator(); slotIterator.hasNext();) {
//                Slot slot = (Slot) slotIterator.next();
//                Vector validInstances = new Vector();
//                Vector allowedClasses = new Vector(slot.getDirectDomain());
//                for (Iterator iter = allowedClasses.iterator(); iter.hasNext();) {
//                    Cls allowedCls = (Cls) iter.next();
//                    Vector instancesOfAllowedCls = new Vector(allowedCls.getInstances());
//                    for (Iterator iterator = instancesOfAllowedCls.iterator(); iterator.hasNext();) {
//                        Instance instanceOfAllowedCls = (Instance) iterator.next();
//                        if (!validInstances.contains(instanceOfAllowedCls)) {
//                            validInstances.add(instanceOfAllowedCls);
//                        }
//                    }
//                }
//
//                // now look at all the instances we've collected to see if they
//                // are roots
//                // an instance cannot be a root if it the "destination" or value
//                // of the cprel slot we are looking at
//                // an instance that is not a "destination" of any slot is
//                // considered a root instance
//                Vector validFramesClone = (Vector) validInstances.clone();
//                for (Iterator iter = validFramesClone.iterator(); iter.hasNext();) {
//                    Instance instance = (Instance) iter.next();
//                    Vector owlSlotValuesOfInstance = getOwnSlotValues(instance, slot);
//                    for (Iterator iterator = owlSlotValuesOfInstance.iterator(); iterator.hasNext();) {
//                        Object ownSlotValueOfInstance = iterator.next();
//                        if (ownSlotValueOfInstance instanceof Instance) {
//                            Instance instanceValue = (Instance) ownSlotValueOfInstance;
//                            if (validInstances.contains(instanceValue)) {
//                                validInstances.removeElement(instanceValue); // the
//                                                                             // value
//                                                                             // of
//                                                                             // this
//                                                                             // slot
//                                                                             // is
//                                                                             // an
//                                                                             // instance,
//                                                                             // so
//                                                                             // this
//                                                                             // instance
//                                                                             // cannot
//                                                                             // be a
//                                                                             // root
//                            }
//                        }
//                    }
//                }
//                // convert the found instances to artifacts
//                for (Iterator iter = validInstances.iterator(); iter.hasNext();) {
//                    Instance instance = (Instance) iter.next();
//                    Artifact art = findArtifact(instance);
//                    if (art != null) {
//                        rootArts.addElement(art);
//                    }
//                }
//            }
        }
        return rootArts;
    }


    /*
     * (non-Javadoc)
     * 
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#findIncomingRelationshipsInBackEnd(ca.uvic.csr.shrimp.DataBean.Artifact)
     */
    protected Vector findIncomingRelationshipsInBackEnd(Artifact artifact) {
        Vector incomingRels = new Vector();
        if (!(artifact instanceof ProtegeArtifact) || ((ProtegeArtifact) artifact).getFrame() == null) {
			return incomingRels;
		}
        findIncomingInstanceRels(artifact, incomingRels);
        findIncomingTemplateRels(artifact, incomingRels);
        if (UNREIFY_RELATION_INSTANCES) {
            incomingRels.addAll(findIncomingUnreifiedRelations(artifact));
        }
        return incomingRels;
    }

    private void findIncomingTemplateRels(Artifact artifact, Vector incomingRels) {
        Frame frame = ((ProtegeArtifact) artifact).getFrame();
        if (!(frame instanceof Cls)) {
			return; // only clses have template rels
		}
        Collection references = knowledgeBase.getReferences(frame, -1);
        // we are looking for any references to a "value type" slot -
        // the "value type" slot is actually a slot, or property, of all
        // instances
        // of ":SLOT". The "value type" slot is also known as the "allowed
        // classes"
        for (Iterator iter = references.iterator(); iter.hasNext();) {
            Reference reference = (Reference) iter.next();
            Frame refFrame = reference.getFrame();
            Slot refSlot = reference.getSlot();
            if (refFrame instanceof Slot && refSlot.equals(knowledgeBase.getSlot(Model.Slot.VALUE_TYPE))) {
                Slot slot = (Slot) refFrame;
                boolean shouldImportSlot = shouldImportSlotAsRelationshipCheckFiltered(slot, SUFFIX_SLOT_TEMPLATE_ALLOWED_CLASS);
                if (shouldImportSlot) {
                    ValueType valueType = getSlotValueType(slot);
                    // we don't consider slot value types of "ANY" here
                    if (valueType == ValueType.CLS || valueType == ValueType.INSTANCE) { 
                        Collection allowedValues = slot.getAllowedValues();
                        if (allowedValues.contains(frame)) {
                            Collection directDomain = slot.getDirectDomain();
                            for (Iterator iterator = directDomain.iterator(); iterator.hasNext();) {
                                Cls srcFrame = (Cls) iterator.next();
                                if (shouldImportFrameAsArtifact(srcFrame)) {
                                    Artifact srcArtifact = findArtifact(srcFrame);
                                    if (srcArtifact != null) {
                                        String relType = slotToRelationshipType(slot, SUFFIX_SLOT_TEMPLATE_ALLOWED_CLASS);
                                        Relationship rel = createRel(null, srcArtifact, artifact, relType);
                                        incomingRels.addElement(rel);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * @param artifact
     * @param incomingRels
     */
    private void findIncomingInstanceRels(Artifact artifact, Vector incomingRels) {
        Frame frame = ((ProtegeArtifact) artifact).getFrame();
        Collection references = knowledgeBase.getReferences(frame, -1);
        for (Iterator iter = references.iterator(); iter.hasNext();) {
            Reference reference = (Reference) iter.next();
            Frame refFrame = reference.getFrame();
            Slot refSlot = reference.getSlot();
            if (refFrame != null && shouldImportFrameAsArtifact(refFrame)) {
                if (refSlot != null) {
                    ValueType valueType = getSlotValueType(refSlot);
                    boolean shouldImportSlot = shouldImportSlotAsRelationshipCheckFiltered(refSlot, SUFFIX_SLOT_INSTANCE);
                    if (shouldImportSlot) {
                        // Note: we don't consider slot value types of "ANY" here
                        if (valueType == ValueType.CLS || valueType == ValueType.INSTANCE) { 
                            Artifact srcArtifact = findArtifact(refFrame);
                            if (srcArtifact != null) {
                                String relType = slotToRelationshipType(refSlot, SUFFIX_SLOT_INSTANCE);
                                ProtegeRelationship rel = (ProtegeRelationship) createRel(null, srcArtifact, artifact, relType);
                                incomingRels.addElement(rel);
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#findOutgoingRelationshipsInBackEnd(ca.uvic.csr.shrimp.DataBean.Artifact)
     */
    protected Vector findOutgoingRelationshipsInBackEnd(Artifact artifact) {
        ProtegeArtifact protArt = (ProtegeArtifact) artifact;
        Frame srcFrame = protArt.getFrame();
        if (srcFrame == null) {
			return new Vector();
		}

        Vector outgoingRels = new Vector();

        findOutgoingInstanceRels(protArt, srcFrame, outgoingRels);

        // Grab any template slots that have this as a possible parent or value
        // and add it to the list
        // Note: This frame already has an artifact so doesn't need to be
        // checked for shouldImport.
        if (srcFrame instanceof Cls) {
            outgoingRels.addAll(findOutgoingTemplateRels(protArt, srcFrame));
        }

        /*
         * //temporary for flow diagrams, add terminal to terminal links for
         * (Iterator iter = ttls.iterator(); iter.hasNext();) { Relationship ttl =
         * (Relationship) iter.next(); if
         * (ttl.getArtifacts().elementAt(0).equals(artifact)) {
         * ret.addElement(ttl); } }
         */

        //check if any unreified rels for this artifact
        if (UNREIFY_RELATION_INSTANCES) {
            outgoingRels.addAll(findOutgoingUnreifiedRelations(artifact));
        }

        return outgoingRels;
    }

    protected Vector findOutgoingUnreifiedRelations(Artifact artifact) {
        if (!relsUnreified) {
            unreifyRelationInstances();
        }
        Collection unreifiedRels = (Collection) artifactToOutgoingUnreifiedRels.get(artifact);
        return unreifiedRels == null ? new Vector() : new Vector(unreifiedRels);
    }

    protected Vector findIncomingUnreifiedRelations(Artifact artifact) {
        if (!relsUnreified) {
            unreifyRelationInstances();
        }
        Collection unreifiedRels = (Collection) artifactToIncomingUnreifiedRels.get(artifact);
        return unreifiedRels == null ? new Vector() : new Vector(unreifiedRels);
    }

    /**
     * @param protArt
     * @param srcFrame
     * @param outgoingRels
     */
    private void findOutgoingInstanceRels(ProtegeArtifact protArt, Frame srcFrame, Vector outgoingRels) {
        Vector slots = new Vector(getOwnSlots(srcFrame));
        for (Iterator iter = slots.iterator(); iter.hasNext();) {
            Slot slot = (Slot) iter.next();

            // we don't consider slot value types of "ANY" here
            ValueType valueType = getSlotValueType(slot);
            if (shouldImportSlotAsRelationshipCheckFiltered(slot, SUFFIX_SLOT_INSTANCE) && (valueType == ValueType.CLS || valueType == ValueType.INSTANCE)) {
                Vector ownSlotValues = getOwnSlotValues(srcFrame, slot);

                for (Iterator iterator = ownSlotValues.iterator(); iterator.hasNext();) {
                    Object ownSlotValue = iterator.next();

                    // if the slot contains other frames as values...
                    if (ownSlotValue instanceof Frame && shouldImportFrameAsArtifact((Frame) ownSlotValue, true)) {
                        Frame destFrame = (Frame) ownSlotValue;
                        Artifact destArt = findArtifact(destFrame);
                        if (destArt != null) {
                            String relType = slotToRelationshipType(slot, SUFFIX_SLOT_INSTANCE);
                            ProtegeRelationship rel = (ProtegeRelationship) createRel(null, protArt, destArt, relType);
                            outgoingRels.addElement(rel);
                        }
                    }
                }
            }
        }
    }

    private Vector findOutgoingTemplateRels(ProtegeArtifact protArt, Frame frame) {
        Vector ret = new Vector();
        if (!(frame instanceof Cls)) {
			return ret; // only clses have template rels
		}
        Cls cls = (Cls) frame;
        Collection slots = cls.getDirectTemplateSlots();
        for (Iterator iter = slots.iterator(); iter.hasNext();) {
            Slot slot = (Slot) iter.next();
            if (shouldImportSlotAllowedValues) {
                if (shouldImportSlotAsRelationshipCheckFiltered(slot, SUFFIX_SLOT_TEMPLATE_ALLOWED_CLASS)) {
                    // these are the classes that the slot is allowed to have as values
                    Collection allowedValues = slot.getAllowedClses(); 
                    for (Iterator iterator = allowedValues.iterator(); iterator.hasNext();) {
                        Frame allowedValue = (Frame) iterator.next();
                        Artifact art = findArtifact(allowedValue);
                        if (art != null) {
                            String relType = slotToRelationshipType(slot, SUFFIX_SLOT_TEMPLATE_ALLOWED_CLASS);
                            Relationship rel = createRel(null, protArt, art, relType);
                            ret.addElement(rel);
                        }
                    }
                }
            }
            if (shouldImportSlotTemplateValues) {
                if (shouldImportSlotAsRelationshipCheckFiltered(slot, SUFFIX_SLOT_TEMPLATE_VALUE)) {
                    // these are template values for this slot
                    Collection templateValues = cls.getDirectTemplateSlotValues(slot); 
                    for (Iterator iterator = templateValues.iterator(); iterator.hasNext();) {
                        Frame templateValue = (Frame) iterator.next();
                        Artifact art = findArtifact(templateValue);
                        if (art != null) {
                            String relType = slotToRelationshipType(slot, SUFFIX_SLOT_TEMPLATE_VALUE);
                            Relationship rel = createRel(null, protArt, art, relType);
                            ret.addElement(rel);
                        }
                    }
                }
            }
        }

        return ret;
    }

    /*
     * (non-Javadoc)
     * 
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#findArtifactByExternalId(java.lang.Object)
     */
    public Artifact findArtifactByExternalId(Object externalId) {
        if (!(externalId instanceof String)) {
            return null;
        }
        return findArtifact((String)externalId, true, true, true);
    }

    public Artifact findArtifact(Frame frame) {
        return findArtifact(frame, true, true, true);
    }

    protected Artifact findArtifact(Frame frame, boolean checkShouldImportArtType, boolean checkIsInRootClses, boolean checkIsHidden) {
        String externalId = frameToExternalArtifactID(frame);
        return findArtifact(frame, externalId, checkShouldImportArtType, checkIsInRootClses, checkIsHidden);
    }
    
    protected Artifact findArtifact(String externalId, boolean checkShouldImportArtType, boolean checkIsInRootClses, boolean checkIsHidden) {
        Frame frame = knowledgeBase.getFrame(externalId);
        return findArtifact(frame, externalId, checkShouldImportArtType, checkIsInRootClses, checkIsHidden);
    }
    
    protected Artifact findArtifact(Frame frame, String externalId, boolean checkShouldImportArtType, boolean checkIsInRootClses, boolean checkIsHidden) {
        // check if we should be considering this type of frame
        if (frame == null) {
            // do nothing
            return null;
        } else if (shouldImportFrameAsArtifact(frame, checkShouldImportArtType, checkIsInRootClses, checkIsHidden)) {
            Artifact art = null;
            Long internalIdLong = (Long)bufferExternalToInternalIdMap.get(externalId);
            if (internalIdLong != null) {
                art = getArtifactReference (internalIdLong);
            }
            if (art == null) {
                art = createArtifact (externalId);
                if (art != null) {
                    addArtifactReference (art);
                }
            }
            return art;
        }
        return null;
    }

    protected Vector getOwnSlots(Frame frame) {
        Vector ownSlots = (Vector) frameToOwnSlots.get(frame);
        if (ownSlots == null) {
            ownSlots = new Vector(frame.getOwnSlots());
            frameToOwnSlots.put(frame, ownSlots);
        }
        return ownSlots;
    }

    protected Vector getOwnSlotValues(Frame frame, Slot slot) {
        Map slotToOwnSlotValues = (Map) frameToOwnSlotValues.get(frame);
        if (slotToOwnSlotValues == null) {
            slotToOwnSlotValues = new HashMap();
            frameToOwnSlotValues.put(frame, slotToOwnSlotValues);
        }
        Vector slotValues = (Vector) slotToOwnSlotValues.get(slot);
        if (slotValues == null) {
            slotValues = new Vector(frame.getOwnSlotValues(slot));
            slotToOwnSlotValues.put(slot, slotValues);
        }
        return slotValues;
    }

    public boolean shouldImportFrameAsArtifact(Frame frame) {
        return shouldImportFrameAsArtifact(frame, true, true, true);
    }

    /**
     * Determines if a frame should be considered for use in shrimp
     */
    public boolean shouldImportFrameAsArtifact(Frame frame, boolean checkShouldImportArtType) {
        return shouldImportFrameAsArtifact(frame, checkShouldImportArtType, true, true);
    }

    /**
     * Determines if a frame should be considered for use in shrimp
     */
    public boolean shouldImportFrameAsArtifact(Frame frame, boolean checkShouldImportArtType, boolean checkIsInRootClses) {
        return shouldImportFrameAsArtifact(frame, checkShouldImportArtType, checkIsInRootClses, true);
    }

    /**
     * Determines if a frame should be considered for use in shrimp
     */
    public boolean shouldImportFrameAsArtifact(Frame frame, boolean checkShouldImportArtType, boolean checkIsInRootClses, boolean checkIsHidden) {
        boolean shouldImport = isProperFrameType(frame);
        if (shouldImport && checkShouldImportArtType) {
            shouldImport = shouldImport && isProperArtifactType(frame);
        }
        if (shouldImport && checkIsInRootClses) {
            shouldImport = shouldImport && isInRootClses(frame);
        }
        if (shouldImport && checkIsHidden) {
            shouldImport = shouldImport && !isHidden(frame);
        }
        if (shouldImport && !showIncludedFrames && !frame.isSystem()) {
            shouldImport = shouldImport && !frame.isIncluded();
        }

        // a quick hack to make sure that subclasses and instances of
        // :DIRECT-BINARY-RELATIONS don't show up
        if (shouldImport) {
            shouldImport = shouldImport && !getRelationSubClassesAndInstances().contains(frame);
        }

        return shouldImport;
    }

    private Collection getRelationSubClassesAndInstances() {
        if (relationSubClassesAndInstances == null) {
            relationSubClassesAndInstances = new ArrayList();
            Cls clsBinaryRelation = knowledgeBase.getCls(":DIRECTED-BINARY-RELATION");
            if (clsBinaryRelation != null) {
                relationSubClassesAndInstances.addAll(clsBinaryRelation.getSubclasses());
                relationSubClassesAndInstances.addAll(clsBinaryRelation.getInstances());
            }
        }
        return relationSubClassesAndInstances;
    }

    protected boolean isHidden(Frame frame) {
        boolean isHidden = false;
        //in Jambalaya's case, we also don't want to show all the system
        // classes, except :THING
        String artType = frameToArtifactType(frame);
        if (artType.equals(SYSTEM_CLASS_ART_TYPE)) {
            isHidden = !knowledgeBase.getRootCls().equals(frame);
        } else if (artType.equals(SYSTEM_META_CLASS_ART_TYPE)) {
            isHidden = true;
        } else {
            isHidden = !frame.isVisible();
        }
        return isHidden;
    }

    protected boolean isInRootClses(Frame frame) {
        boolean b = true;
        if (frame instanceof Cls) {
            Cls cls = (Cls) frame;
            boolean isSubClsOfRootCls = false;
            for (Iterator iter = rootClses.iterator(); iter.hasNext() && !isSubClsOfRootCls;) {
            	Frame rootFrame = (Frame) iter.next();
            	if (rootFrame instanceof Cls) {
					Cls rootCls = (Cls) rootFrame;
	                if (cls == rootCls || cls.hasSuperclass(rootCls)) {
	                    isSubClsOfRootCls = true;
	                } else if (cls.getName().equals(rootCls.getName())) {
	                    isSubClsOfRootCls = true;
	                }
				}
            }
            b = isSubClsOfRootCls;
        } else if (frame instanceof Instance) {
            Instance instance = (Instance) frame;
            boolean isInstanceOfRootCls = false;
            for (Iterator iter = rootClses.iterator(); iter.hasNext() && !isInstanceOfRootCls;) {
            	Frame rootFrame = (Frame) iter.next();
                if (rootFrame == instance) {
                	isInstanceOfRootCls = true;
                } else if (rootFrame instanceof Cls) {
					Cls rootCls = (Cls) rootFrame;
	                if (instance.hasType(rootCls)) {
	                    isInstanceOfRootCls = true;
	                }
				}
            }
            b = isInstanceOfRootCls;
        }
        return b;
    }

    private boolean isProperFrameType(Frame frame) {
        return !(frame instanceof Slot) && !(frame instanceof Facet);
    }

    private boolean isProperArtifactType(Frame frame) {
        return !isArtTypeFiltered(frameToArtifactType(frame));
    }

    /**
     * Returns the type of the artifact that will represent the given frame.
     */
    protected String frameToArtifactType(Frame frame) {
        String type = (String) frameToArtifactTypeMap.get(frame);
        if (type == null) {
            if (frame instanceof Cls) {
                if (frame.isSystem()) {
                    if (((Cls) frame).isMetaCls()) {
                        type = SYSTEM_META_CLASS_ART_TYPE;
                    } else {
                        type = SYSTEM_CLASS_ART_TYPE;
                    }
                } else if (((Cls) frame).isClsMetaCls()) {
                    type = META_CLASS_ART_TYPE;
                } else {
                    type = CLASS_ART_TYPE; //: " + getBrowserText(frame);
                }
            } else if (frame instanceof Instance) {
                //Cls directType = ((Instance)frame).getDirectType();
                //type = "Instance of: " + getBrowserText(directType);
                type = INSTANCE_ART_TYPE;
            } else {
                type = UNKNOWN_ART_TYPE;
            }
            frameToArtifactTypeMap.put(frame, type);
        }
        return type;
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.DataBean#getArtifactIcon(ca.uvic.csr.shrimp.DataBean.Artifact)
     */
    public Icon getArtifactIcon(Artifact artifact) {
        Icon icon = null;
        Frame frame = ((ProtegeArtifact) artifact).getFrame();
        if (frame != null) {
            icon = frame.getIcon();
        }
        return icon;
    }

    /**
     * Returns the type of the relationship that will represent the given slot.
     */
    protected String slotToRelationshipType(Slot slot, String suffix) {
        String type = getBrowserText(slot);
        if (slot.equals(knowledgeBase.getSlot(Model.Slot.DIRECT_SUBCLASSES))) {
            type = DIRECT_SUBCLASS_SLOT_TYPE;
        } else if (slot.equals(knowledgeBase.getSlot(Model.Slot.DIRECT_INSTANCES))) {
            type = DIRECT_INSTANCE_SLOT_TYPE;
        } else {
            type += suffix;
        }
        return type;
    }
    
    /**
     * Converts from a relationship type (with suffix) back to the slot that it is derived from.
     * @param relType
     * @return
     */
    protected Slot relationshipTypeToSlot (String relType) {
        Slot slot = null;
        String baseRelType = relType;
        if (relType.lastIndexOf(SUFFIX_SLOT_INSTANCE) != -1) {
            baseRelType = relType.substring(relType.lastIndexOf(SUFFIX_SLOT_INSTANCE));
        } else if (relType.lastIndexOf(SUFFIX_SLOT_REIFIED_REL) != -1) {
            baseRelType = relType.substring(relType.lastIndexOf(SUFFIX_SLOT_REIFIED_REL));
        } else if (relType.lastIndexOf(SUFFIX_SLOT_TEMPLATE_ALLOWED_CLASS) != -1) {
            baseRelType = relType.substring(relType.lastIndexOf(SUFFIX_SLOT_TEMPLATE_ALLOWED_CLASS));
        } else if (relType.lastIndexOf(SUFFIX_SLOT_TEMPLATE_VALUE) != -1) {
            baseRelType = relType.substring(relType.lastIndexOf(SUFFIX_SLOT_TEMPLATE_VALUE));
        } else if (relType.equals(DIRECT_INSTANCE_SLOT_TYPE)) {
            return knowledgeBase.getSlot(Model.Slot.DIRECT_INSTANCES);
        } else if (relType.equals(DIRECT_SUBCLASS_SLOT_TYPE)) {
            return knowledgeBase.getSlot(Model.Slot.DIRECT_SUBCLASSES);
        }
        slot = knowledgeBase.getSlot(baseRelType);
        return slot;
        
    }

    /**
     * Determines if a slot should be considered for use in shrimp.
     * 
     * @param slot
     *            The slot to consider.
     * @param checkRelType
     *            Check whether or not the type of the relationship representing
     *            this slot should be checked.
     */
    protected boolean shouldImportSlotAsRelationshipCheckFiltered(Slot slot, String suffix) {
        String relType = slotToRelationshipType(slot, suffix);
        if (isRelTypeFiltered(relType)) {
            return false;
        }

        return shouldImportSlotAsRelationship(slot);
    }

    protected boolean slotIsRightSystemType(String baseRelType) {
        boolean rightSystemType = baseRelType.equals(DIRECT_INSTANCE_SLOT_TYPE) || baseRelType.equals(DIRECT_SUBCLASS_SLOT_TYPE) || baseRelType.equals(":FROM") || baseRelType.equals(":TO"); // :TO
        return rightSystemType;
    }

    protected boolean shouldImportSlotAsRelationship(Slot slot) {
        // import only system slots of the right type
        String baseRelType = slotToRelationshipType(slot, "");
        if (slot.isSystem()) {
            boolean rightSystemType = slotIsRightSystemType(baseRelType);
            if (!rightSystemType) {
                return false;
            }
        } else {
            if (!showIncludedFrames && slot.isIncluded()) {
                return false;
            }           
        }

        // we only want slots that have a value type of class, instance, or any
        boolean slotHasValidRange = slotHasValidRange(slot);
        if (!slotHasValidRange) {
            return false;
        }

        boolean hasOneValidCls = slotHasOneValidCls(slot);
        if (!hasOneValidCls) {
            return false;
        }

        if (isHidden(slot)) {
            return false;
            //System.err.println("importing hidden slot!: " +
            // getBrowserText(slot));
        }

        //if we've made it here this slot is valid and should be imported
        return true;
    }

    protected boolean slotHasValidRange(Slot slot) {
        ValueType slotValueType = getSlotValueType(slot);
        if (slotValueType != ValueType.CLS && slotValueType != ValueType.INSTANCE && slotValueType != ValueType.ANY) {
            return false;
        }
        return true;
    }

    protected boolean slotHasOneValidCls(Slot slot) {
        // see if there is at least one allowed class that is acceptable
        boolean hasOneValidAllowedCls = false;
        for (Iterator iter = slot.getAllowedClses().iterator(); iter.hasNext() && !hasOneValidAllowedCls;) {
            Cls allowedCls = (Cls) iter.next();
            if (shouldImportFrameAsArtifact(allowedCls, true) || validSlotSuperClsesAndTypes.contains(allowedCls)) {
                hasOneValidAllowedCls = true;
            }
        }
        if (!hasOneValidAllowedCls) {
            return false;
        }

        // see if there is at least one template slot class that is acceptable
        boolean hasOneValidTemplateSlotCls = false;
        for (Iterator iter = slot.getDirectDomain().iterator(); iter.hasNext() && !hasOneValidTemplateSlotCls;) {
            Cls templateSlotCls = (Cls) iter.next();
            if (shouldImportFrameAsArtifact(templateSlotCls, true) || validSlotSuperClsesAndTypes.contains(templateSlotCls)) {
                hasOneValidTemplateSlotCls = true;
            }
        }
        if (!hasOneValidTemplateSlotCls) {
            return false;
        }
        return true;
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#addRelationshipToBackEnd(ca.uvic.csr.shrimp.DataBean.Relationship)
     */
    protected boolean addRelationshipToBackEnd(Relationship relationship) {
        // TODO implement ProtegeDataBean.addRelationshipToBackEnd
        return false;
    }

    /**
     * Removes the relationship from protege
     */
    protected boolean removeRelationshipFromBackEnd(Relationship relationship) {
        // TODO implement ProtegeDataBean.removeRelationshipFromBackEnd
        boolean childRemoved = false;

        return childRemoved;
    }

    /*
     * (non-Javadoc)
     * 
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#getParents(ca.uvic.csr.shrimp.DataBean.Artifact,
     *      java.lang.String[])
     */
    public Vector getParents(Artifact artifact, String[] cprels) {
        if (isFasterWayToCount(artifact, cprels)) {
            List cprelsList = Arrays.asList(cprels);
            Frame frame = ((ProtegeArtifact) artifact).getFrame();
            Vector potentialParentFrames = new Vector();
            if (frame instanceof Cls) {
                Cls cls = (Cls) frame;
                if (cprelsList.contains(DIRECT_SUBCLASS_SLOT_TYPE)) {
                    Collection allDirectSuperClasses = cls.getDirectSuperclasses();
                    for (Iterator iter = allDirectSuperClasses.iterator(); iter.hasNext();) {
                        Cls superCls = (Cls) iter.next();
                        if (isValidSubClass(cls, superCls)) {
                            potentialParentFrames.add(superCls);
                        }
                    }
                }
                if (cprelsList.contains(DIRECT_INSTANCE_SLOT_TYPE)) {
                    Collection directTypes = cls.getDirectTypes();
                    for (Iterator iter = directTypes.iterator(); iter.hasNext();) {
                        Cls type = (Cls) iter.next();
                        if (!type.equals(frame)) {
                            //TODO should we get the direct types of a cls?
                            //Artifact typeArt = findArtifact (type, true,
                            // false, true);
                            //if (typeArt != null) {
                            //System.out.println("\nvalid direct type of '" +
                            // getBrowserText(cls) + "' is '" +
                            // getBrowserText(type));
                            //}
                            potentialParentFrames.add(type);
                        }
                    }
                }
            } else if (frame instanceof SimpleInstance) {
                SimpleInstance simpleInstance = (SimpleInstance) frame;
                if (cprelsList.contains(DIRECT_INSTANCE_SLOT_TYPE)) {
                    Collection directTypes = simpleInstance.getDirectTypes();
                    for (Iterator iter = directTypes.iterator(); iter.hasNext();) {
                        Cls type = (Cls) iter.next();
                        if (!type.equals(frame)) {
                            potentialParentFrames.add(type);
                        }
                    }
                }

            }
            Vector parents = new Vector();
            for (Iterator iter = potentialParentFrames.iterator(); iter.hasNext();) {
                Frame potentialParent = (Frame) iter.next();
                Artifact parentArt = findArtifact(potentialParent, true, true, true);
                // shouldn't be duplicates, takes too long to do the .contains
                if (parentArt != null /* && !parents.contains(parentArt) */) { 
                    parents.add(parentArt);
                }
            }
            //System.out.println("parents of '" + getBrowserText(frame) + "' =
            // " + parents);
            return parents;
        }
        return super.getParents(artifact, cprels);
    }

    /*
     * (non-Javadoc)
     * 
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#getChildren(ca.uvic.csr.shrimp.DataBean.Artifact,
     *      java.lang.String[])
     */
    public Vector getChildren(Artifact artifact, String[] cprels) {
        if (isFasterWayToCount(artifact, cprels)) {
            List cprelsList = Arrays.asList(cprels);
            Frame frame = ((ProtegeArtifact) artifact).getFrame();
            Vector potentialChildFrames = new Vector();
            if (frame instanceof Cls) {
                Cls cls = (Cls) frame;
                if (cprelsList.contains(DIRECT_SUBCLASS_SLOT_TYPE)) {
                    Collection allDirectSubClasses = cls.getDirectSubclasses();
                    for (Iterator iter = allDirectSubClasses.iterator(); iter.hasNext();) {
                        Cls subCls = (Cls) iter.next();
                        if (isValidSubClass(cls, subCls)) {
                            potentialChildFrames.add(subCls);
                        }
                    }
                }
                if (cprelsList.contains(DIRECT_INSTANCE_SLOT_TYPE)) {
                    Collection directInstances = ((Cls) frame).getDirectInstances();
                    for (Iterator iter = directInstances.iterator(); iter.hasNext();) {
                        Instance inst = (Instance) iter.next();
                        if (!inst.equals(frame)) {
                            potentialChildFrames.add(inst);
                        }
                    }
                }
            } else if (frame instanceof SimpleInstance) {
                // no children of simple instances
            }
            Vector children = new Vector();
            //int i = 0;
            for (Iterator iter = potentialChildFrames.iterator(); iter.hasNext();) {
                // System.out.println(i++);
                Frame potentialChild = (Frame) iter.next();
                Artifact childArt = findArtifact(potentialChild, true, false, true);
                if (childArt != null /* && !children.contains(childArt) */) { //shouldn't be duplicates
                    children.add(childArt);
                }
            }
            return children;
        }
        return super.getChildren(artifact, cprels);
    }

    protected boolean isValidSubClass(Cls cls, Cls subCls) {
        return !cls.equals(subCls);
    }

    protected boolean isFasterWayToCount(Artifact artifact, String[] cprels) {
        List cprelsList = Arrays.asList(cprels);
        boolean countInstances = cprelsList.contains(DIRECT_INSTANCE_SLOT_TYPE);
        boolean countSubClasses = cprelsList.contains(DIRECT_SUBCLASS_SLOT_TYPE);
        Frame frame = ((ProtegeArtifact) artifact).getFrame();
        boolean rightType = (frame instanceof SimpleInstance) || (frame instanceof Cls);
        boolean fasterWayToCount = rightType && (cprelsList.size() == 2 && countInstances && countSubClasses) || (cprelsList.size() == 1 && (countInstances || countSubClasses));
        return fasterWayToCount;
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#getChildrenCount(Artifact,
     *      String []) Overriden to make a bit faster for large knowledge bases.
     *      We can get count of children quicker if we don't pull in all
     *      relationships.
     */
    public int getChildrenCount(Artifact artifact, String[] cprels) {
        if (cprels.length == 0) {
			return 0;
		}
        int childrenCount = 0;
        Frame frame = ((ProtegeArtifact) artifact).getFrame();
        if (isFasterWayToCount(artifact, cprels)) {
            List cprelsList = Arrays.asList(cprels);
            boolean countInstances = cprelsList.contains(DIRECT_INSTANCE_SLOT_TYPE);
            boolean countSubClasses = cprelsList.contains(DIRECT_SUBCLASS_SLOT_TYPE);
            childrenCount = getFasterChildrenCount(frame, countInstances, countSubClasses);
            // add cprels to attribute name because number of children depends on the given cprels
            artifact.setAttribute(AttributeConstants.ORD_ATTR_NUM_CHILDREN + cprelsToKey(cprels), new Integer(childrenCount)); 
        } else {
            childrenCount = super.getChildrenCount(artifact, cprels);
        }
        return childrenCount;
    }

    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#getDescendentsCount(ca.uvic.csr.shrimp.DataBean.Artifact,
     *      java.lang.String[]) Overriden to make a bit faster for large
     *      knowledge bases. We can get count of descendents quicker if we don't
     *      pull in all relationships.
     */
    /**
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#getDescendentsCount(ca.uvic.csr.shrimp.DataBean.Artifact,
     *      java.lang.String[], boolean)
     */
    public int getDescendentsCount(Artifact artifact, String[] cprels, boolean countArtifactMultipleTimes) {
        if (cprels.length == 0) {
            return 0;
        }
        Frame frame = ((ProtegeArtifact) artifact).getFrame();
        int numDes = 0;
        if (isFasterWayToCount(artifact, cprels)) {
            List cprelsList = Arrays.asList(cprels);
            boolean countInstances = cprelsList.contains(DIRECT_INSTANCE_SLOT_TYPE);
            numDes = getFasterDescendentsCountRecursive(frame, countInstances, countArtifactMultipleTimes);
            // add cprels to attribute name because number of descendents depends on the given cprels
            artifact.setAttribute(AttributeConstants.ORD_ATTR_NUM_DESCENDENTS + cprelsToKey(cprels), new Integer(numDes)); 
        } else {
            numDes = getDescendentsCountRecursive(artifact, cprels, new HashSet(), new HashSet(), countArtifactMultipleTimes);
        }
        return numDes;
    }

    public int getFasterChildrenCount(Frame frame, boolean countInstances, boolean countSubClasses) {
        int childrenCount = 0;
        if (frame instanceof Cls) {
            Cls cls = (Cls) frame;
            if (countSubClasses) {
                Collection subClses = cls.getDirectSubclasses();
                for (Iterator iter = subClses.iterator(); iter.hasNext();) {
                    Cls subCls = (Cls) iter.next();
                    if (shouldImportFrameAsArtifact(subCls, true)) {
                        childrenCount++;
                    }
                }
            }
            if (countInstances /* && (CONSIDER_METACLASS_INSTANCES || !cls.isMetaCls()) */) {
                Collection instances = cls.getDirectInstances();
                for (Iterator iter = instances.iterator(); iter.hasNext();) {
                    Instance instance = (Instance) iter.next();
                    if (shouldImportFrameAsArtifact(instance, true)) {
                        childrenCount++;
                    }
                }
            }
        }
        return childrenCount;
    }
    
    protected void getFasterDescendentsInstancesCountRecursive (Frame frame, boolean countInstances, Set seenSoFar, Set pathSoFar, boolean countArtifactMultipleTimes, int [] instanceCount, int [] subclassCount) {
        if (frame instanceof Cls) {
            Cls cls = (Cls) frame;
            if ((!countArtifactMultipleTimes && seenSoFar.contains(frame)) || pathSoFar.contains(frame)) {
                System.err.println("on path already!: " + frame);
                return; // prevent infinite loops
            }
            pathSoFar.add(frame);
            seenSoFar.add(frame);
            Collection directInstances = cls.getDirectInstances();
            Collection directSubClasses = cls.getDirectSubclasses();
            if (countInstances) {
	            for (Iterator iter = directInstances.iterator(); iter.hasNext();) {
	                Instance instance = (Instance) iter.next();
	                if (shouldImportFrameAsArtifact(instance, true, false, true)) {
	                    if ((countArtifactMultipleTimes || !seenSoFar.contains(instance)) && !pathSoFar.contains(instance)) {
	                        instanceCount[0]++;
	                    }
	                }
	            }
            }
            for (Iterator iter = directSubClasses.iterator(); iter.hasNext();) {
                Cls subCls = (Cls) iter.next();
                if (shouldImportFrameAsArtifact(subCls, true, false, true)) {
	                if ((countArtifactMultipleTimes || !seenSoFar.contains(subCls)) && !pathSoFar.contains(subCls)) {
                        subclassCount[0]++;
	                    getFasterDescendentsInstancesCountRecursive(subCls, countInstances, seenSoFar, pathSoFar, countArtifactMultipleTimes, instanceCount, subclassCount);
	                }                        
                }
            }  
            pathSoFar.remove(frame);
        }
    }
    
    
    protected int getFasterDescendentsCountRecursive(Frame frame, boolean countInstances, boolean countArtifactMultipleTimes) {
        int count = 0;
        int [] instancesCount = new int [] {0}; // using arrays so can pass by reference
        int [] subclassesCount = new int [] {0}; 
        getFasterDescendentsInstancesCountRecursive(frame, countInstances, new HashSet(), new HashSet(), countArtifactMultipleTimes, instancesCount, subclassesCount);
        if (countInstances) {
            count += instancesCount[0];
        }
        count += subclassesCount[0];
        return count;
    }

    protected void getDescendentsRecursive(Frame frame, boolean countInstances, boolean countSubClasses, Set seenSoFar, Set pathSoFar, boolean countArtifactMultipleTimes, Set descendents) {
        if (ProgressDialog.isCancelled()) {
			return;
		}
        if (frame instanceof Cls) {
            Cls cls = (Cls) frame;
            if ((!countArtifactMultipleTimes && seenSoFar.contains(frame)) || pathSoFar.contains(frame)) {
                //System.err.println("on path already!: " + frame);
                return; // prevent infinite loops
            }
            pathSoFar.add(frame);
            seenSoFar.add(frame);
            if (countSubClasses) {
                Collection subClses = cls.getDirectSubclasses();
                for (Iterator iter = subClses.iterator(); iter.hasNext() && !ProgressDialog.isCancelled();) {
                    Cls subCls = (Cls) iter.next();
                    Artifact artifact = findArtifact(subCls, true, false, true);
                    if (artifact != null) {
                        if ((countArtifactMultipleTimes || !seenSoFar.contains(subCls)) && !pathSoFar.contains(subCls)) {
                            descendents.add(artifact);
                            ProgressDialog.setNote("found " + descendents.size() + (descendents.size() == 1 ? " item" : " items") + " so far ...");
                            getDescendentsRecursive(subCls, countInstances, countSubClasses, seenSoFar, pathSoFar, countArtifactMultipleTimes, descendents);
                        }
                    }
                }
            }
            if (countInstances /*
                                * && (CONSIDER_METACLASS_INSTANCES ||
                                * !cls.isMetaCls())
                                */) {
                Collection instances = cls.getDirectInstances();
                for (Iterator iter = instances.iterator(); iter.hasNext() && !ProgressDialog.isCancelled();) {
                    Instance instance = (Instance) iter.next();
                    Artifact artifact = findArtifact(instance, true, false, true);
                    if (artifact != null) {
                        if ((countArtifactMultipleTimes || !seenSoFar.contains(instance)) && !pathSoFar.contains(instance)) {
                            descendents.add(artifact);
                            ProgressDialog.setNote("found " + descendents.size() + (descendents.size() == 1 ? " item" : " items") + " so far ...");
                            getDescendentsRecursive(instance, countInstances, countSubClasses, seenSoFar, pathSoFar, countArtifactMultipleTimes, descendents);
                        }
                    }
                }
            }
            pathSoFar.remove(frame);
        }
    }

    /** 
     * @return The Protege knowledge base wrapped by this data bean.
     */
    public KnowledgeBase getKnowledgeBase() {
        return knowledgeBase;
    }
    
    /*
     *  (non-Javadoc)
     * @see ca.uvic.csr.shrimp.DataBean.DataBean#getOrdinalAttrValues(java.lang.String[], boolean, boolean, int)
     */
    public Map getOrdinalAttrValues(String[] cprels, boolean lookInBackEnd, boolean inverted, int targetType) {
        Map values = super.getOrdinalAttrValues(cprels, lookInBackEnd, inverted, targetType);
        
        if (targetType != ARTIFACT_TYPE) {
            return values;
        }

        // we need to add a few more attributes for protege from the metrics
        Vector artifacts = getArtifacts(lookInBackEnd);

        //warn the user before doing this if many artifacts?

        SortedSet instancesCounts = new TreeSet();
        SortedSet inheritorCounts = new TreeSet();
        SortedSet depthCounts = new TreeSet();
        SortedSet strahlerCounts = new TreeSet();

        for (Iterator iter = artifacts.iterator(); iter.hasNext();) {
            ProtegeArtifact art = (ProtegeArtifact) iter.next();

            //get from PM
            Integer instances = new Integer(metrics.getNumInstances(art.getFrame()));
            //now set it
            if (instances != null) {
                art.setAttribute(ProtegeArtifact.INSTANCES_KEY, instances);
                instancesCounts.add(instances);
            }

            Integer inheritors = new Integer(metrics.getNumInheritors(art.getFrame()));
            if (inheritors != null) {
                art.setAttribute(ProtegeArtifact.INHERITORS_KEY, inheritors);
                inheritorCounts.add(inheritors);
            }

            Integer strahler = new Integer(metrics.getStrahler(art.getFrame()));
            if (strahler != null) {
                art.setAttribute(ProtegeArtifact.STRAHLER_KEY, strahler);
                strahlerCounts.add(strahler);
            }

            Integer depth = new Integer(metrics.getTreeDepth(art.getFrame()));
            if (depth != null) {
                art.setAttribute(ProtegeArtifact.DEPTH_KEY, depth);
                depthCounts.add(depth);
            }
        }
        values.put(ProtegeArtifact.INSTANCES_KEY, instancesCounts);
        values.put(ProtegeArtifact.INHERITORS_KEY, inheritorCounts);
        values.put(ProtegeArtifact.DEPTH_KEY, depthCounts);
        values.put(ProtegeArtifact.STRAHLER_KEY, strahlerCounts);
        return values;
    }

    /* **** NOTE: the id of a protege node is the same as the name of the 
     * frame that it represents, NOT the frame's id.
     * The reason for this is that for projects stored in a database the
     * frame id can change from session to session, whereas the frame name 
     * stays the same. One issue that arises here, is that the frame name 
     * can be changed by the user while editing a project.
     */
    public static String frameToExternalArtifactID(Frame frame) {
        return frame.getName();
    }

    /**
     * 
     * @param frame
     * @return The name to be given to the artifact associated with the given frame.
     */
    public static String frameToArtifactName(Frame frame) {
        return getBrowserText(frame);
    }

    /* (non-Javadoc)
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#getDefaultGroupForRelationshipType(java.lang.String)
     */
    public String getDefaultGroupForRelationshipType(String relType) {
        if (relTypeToGroupMap.isEmpty()) {
            createGroupToRelTypeMappings();
        }
        String groupName = (String) relTypeToGroupMap.get(relType);
        if (groupName == null) {
            System.err.println("Warning! No group for relationship type: \"" + relType + "\"");
        }
        return groupName;
    }

    protected void createGroupToRelTypeMappings() {
        relTypeToGroupMap.clear();
        relTypeToGroupMap.put(DIRECT_INSTANCE_SLOT_TYPE, GROUP_NAME_CLASS_INSTANCE_HIERARCHY);
        relTypeToGroupMap.put(DIRECT_SUBCLASS_SLOT_TYPE, GROUP_NAME_CLASS_INSTANCE_HIERARCHY);
        Collection slots = knowledgeBase.getSlots();
        for (Iterator iter = slots.iterator(); iter.hasNext();) {
            Slot slot = (Slot) iter.next();
            if (shouldImportSlotAsRelationship(slot)) {
                String baseRelType = slotToRelationshipType(slot, "");
                if (baseRelType.equals(DIRECT_INSTANCE_SLOT_TYPE) || baseRelType.equals(DIRECT_SUBCLASS_SLOT_TYPE)) {
                    continue;
                }
                if (slot.isSystem()) {
                    relTypeToGroupMap.put(baseRelType, GROUP_NAME_SYSTEM);
                } else {
                    relTypeToGroupMap.put(baseRelType + SUFFIX_SLOT_INSTANCE, GROUP_NAME_SLOT_INSTANCE);
                    //relTypeToGroupMap.put(baseRelType + SUFFIX_SLOT_TEMPLATE_VALUE, GROUP_NAME_SLOT_TEMPLATE_VALUE);
                    relTypeToGroupMap.put(baseRelType + SUFFIX_SLOT_TEMPLATE_ALLOWED_CLASS, GROUP_NAME_SLOT_TEMPLATE_ALLOWED_CLASS);
                }
            }
        }
        createGroupToRelTypeMappingsUnreified();
    }

    /**
     * 
     */
    protected void createGroupToRelTypeMappingsUnreified() {
        //add relationship types for reified relationsips
        if (UNREIFY_RELATION_INSTANCES) {
            Cls clsBinaryRelation = knowledgeBase.getCls(":DIRECTED-BINARY-RELATION");
            if (clsBinaryRelation != null) {
                Collection relationSubClasses = clsBinaryRelation.getSubclasses();
                Collection relationInstances = clsBinaryRelation.getInstances();
                relationInstances.addAll(relationSubClasses);
                Slot toSlot = knowledgeBase.getSlot(":TO");
                Slot fromSlot = knowledgeBase.getSlot(":FROM");
                for (Iterator iter = relationInstances.iterator(); iter.hasNext();) {
                    Instance relationInstance = (Instance) iter.next();
                    Instance toInstance = (Instance) relationInstance.getOwnSlotValue(toSlot);
                    Instance fromInstance = (Instance) relationInstance.getOwnSlotValue(fromSlot);
                    if (toInstance != null && fromInstance != null) {
                        Artifact toArtifact = findArtifact(toInstance);
                        Artifact fromArtifact = findArtifact(fromInstance);
                        if (toArtifact != null && fromArtifact != null) {
                            String relType = reifiedRelToRelType(relationInstance);
                            relTypeToGroupMap.put(relType + SUFFIX_SLOT_REIFIED_REL, GROUP_NAME_REIFIED_RELS);
                        }
                    }
                }
            }
        }
    }

    // for testing
    public static void main(String[] args) {
        ApplicationAccessor.setApplication(new AbstractShrimpApplication("", ""){});
        //String fileName = "C:\\Program Files\\Protege_2.1_beta_build_199\\examples\\testCount.pprj";
        //String fileName = "C:\\Program Files\\Protege_2.1_beta_build_199\\examples\\Thesaurus.pprj";
        String fileName = "demo\\newspaper\\newspaper.pprj";
        Collection errors = new ArrayList();
        Project project = new Project(fileName, errors);
        System.out.println("errors: " + errors);
        final ProtegeDataBean dataBean = new ProtegeDataBean(project);
        dataBean.setDefaultRootClses();

        Vector leaves = dataBean.getLeafArtifacts(new String[] { "responsible_for" });
        for (Iterator iter = leaves.iterator(); iter.hasNext();) {
            Artifact leaf = (Artifact) iter.next();
            System.out.println(leaf);
        }

        /*
         KnowledgeBase kb = project.getKnowledgeBase();
         Slot hasDataSlot = kb.getSlot("has_data");
         Slot symbolTypeSlot = kb.getSlot("symbol_type_slot_name");
         Cls hasDataSlotCls = kb.getCls("class_with_has_data_slot");
         Collection instancesOfHasDataSlotCls = kb.getInstances(hasDataSlotCls);
         for (Iterator iter = instancesOfHasDataSlotCls.iterator(); iter.hasNext();) {
         Instance instanceOfHasDataSlotCls = (Instance) iter.next();
         Instance hasDataValueInstance = (Instance) instanceOfHasDataSlotCls.getOwnSlotValue(hasDataSlot);
         String symbol = (String) hasDataValueInstance.getOwnSlotValue(symbolTypeSlot);
         Color color = Color.GRAY;
         if (symbol.equals("a")) {
         color = Color.RED; 
         } else if (symbol.equals("b")) {
         color = Color.BLUE;
         } 
         
         // at this point we can associate "color" with instanceOfHasDataSlotCls
         }
         */
    }

    /* (non-Javadoc)
     * @see ca.uvic.csr.shrimp.DataBean.AbstractDataBean#addArtifactToBackEnd(ca.uvic.csr.shrimp.DataBean.Artifact)
     */
    protected boolean addArtifactToBackEnd(Artifact artifact) {
        // TODO implement ProtegeDataBean.addArtifactToBackEnd
        return false;
    }
    
    public void setShowIncludedFrames (boolean showIncludedFrames) {
        this.showIncludedFrames = showIncludedFrames;
    }
    
    public boolean getShowIncludedFrames () {
        return showIncludedFrames;
    }


}
